////////////////////////////////////////////////////////
//Bibliothèque programmation par graph d'état (Stat Machine)
//Contient les objets :
//Stat, Timer, Event, Button, Pulse, DigitalOut, Bistable, EventFiltered
//Contient les fonctions :
//transition,when et after
//A DUGAS CPGE - Lycée Livet
////////////////////////////////////////////////////////
#include "obj_stat.h"
#include <Arduino.h>


//***********************************
//member definition Stat
//***********************************

//Constructeur
Stat::Stat(const char *stat_name, int stat_id) {
  //Variables carractérisant l'état
  m_entry = false;  //entry_, do_, exit_ ne peuvent pas être a true simultanément
  m_do = false;
  m_exit = false;
  m_stattime = 0;
  completion_event = false;
  //Log
  log = true;
  m_name = String(stat_name);
  id = stat_id;
};

String Stat::name() const{
  return m_name;
}

//Variables carractérisant l'état
bool Stat::entry_() const{
  return m_entry;
};
bool Stat::do_() const{
  return m_do;
};
bool Stat::exit_() const{
  return m_exit;
};
bool Stat::stat() const{
  return (m_entry || m_do);
};
bool Stat::interne_stat() const{
  return m_do;
};


//Activation, Désactivation

Stat *Stat::activate() {
  m_stattime = millis();
  completion_event = false;
  if (!stat()) {
    m_entry = true;
    _logln("Activation de l'etat : " + m_name + " (Id:" + String(id) + ") a " + String(m_stattime) + " ms.", true);
  } else {
    m_entry = true;
    m_exit = true;
    _logln("Re-Activation de l'etat : " + m_name + " (Id:" + String(id) + ")", true);
    _logln("  La date d'activation de l etat est maintenant :" + String(m_stattime) + " ms.");
  }

  return this;  //retourne un pointeur sur l'état actif
};
void Stat::desactivate() {
  completion_event = false;
  if (interne_stat()) {
    m_entry = false;
    m_do = false;
    m_exit = true;
    _logln("Desactivation de l'etat : " + m_name + " (Id:" + String(id) + ") a " + String(millis()) + " ms.", true);
  } else {
    _logln("ATTENTION : Tentative de desactivation d'un etat non actif : " + m_name + " (Id:" + String(id) + ") a " + String(millis()) + " ms.", true);
  }
};

//After
bool Stat::after(unsigned long delay_ms) {  //Vaut true apres un délais imposée depuis l'activation de l'état
  //Serial.println(delay_ms);
  if ((millis() - m_stattime) < delay_ms)
    return false;
  else
    return interne_stat();
};

//After resetStattime
void Stat::resetStattime() {  //Réinitialise m_stattime
  m_stattime = millis();
};

//get_stattime
unsigned long Stat::getStattime() {
  return m_stattime;
}

//Mise à jour
bool Stat::update() {  //Retourne true si il y a une évolution et false sinon
  if (m_entry) {
    m_entry = false;
    m_do = true;
    return true;
  }
  if (m_exit) {
    m_exit = false;
    return true;
  }
  return false;
};

//_log
void Stat::_log(String message, bool titre) {  //Equivalent à println  si log=true (ne fait rien sinon)
  if (log) {
    if (titre)
      Serial.print(F("Stat : "));

    Serial.print(message);
  }
};

void Stat::_logln(String message, bool titre) {  //Equivalent à println  si log=true (ne fait rien sinon)
  if (log) {
    if (titre)
      Serial.print(F("Stat : "));
    Serial.println(message);
  }
};

//*************************************************
//Transitions
//*************************************************
boolean transition(Stat *&pActive,  Stat &from,  Stat &to, boolean event, boolean guard) {  //event: éveinement, guard : condition de garde, from : état avant, to : état après, pActive : pointeur sur l'état actif
  if (!from.interne_stat())                                                                           //l'état avant n'est pas actif
    return false;
  if (!(event && guard))  //la condition n'est pas vérifié
    return false;
//cas de la transition franchissable
#if LOG_TRANSITION
  Serial.println();
  Serial.print(F("-> transition de "));
  Serial.print(from.name());
  Serial.print(F(" vers "));
  Serial.println(to.name());
#endif
  from.desactivate();
  pActive = to.activate();
  return true;
}

boolean completion(Stat *&pActive,  Stat &from,  Stat &to, boolean guard) {  //event: évènement, guard : condition de garde, from : état avant, to : état après, pActive : pointeur sur l'état actif
  if (!from.interne_stat())                                                            //l'état avant n'est pas actif
    return false;
  if (!(from.completion_event && guard))  //la condition n'est pas vérifié
    return false;
//cas de la transition franchissable
#if LOG_TRANSITION
  Serial.println();
  Serial.print(F("-> completion de "));
  Serial.print(from.name());
  Serial.print(F(" vers "));
  Serial.println(to.name());
#endif
  from.desactivate();
  pActive = to.activate();
  return true;
}

boolean when(Stat *&pActive,  Stat &from,  Stat &to, boolean condition, boolean guard) {
  if (!from.interne_stat())  //l'état avant n'est pas actif
    return false;
  if (!(condition && guard))  //la condition n'est pas vérifié
    return false;
//cas de la transition franchissable
#if LOG_TRANSITION
  Serial.println();
  Serial.print(F("-> when de "));
  Serial.print(from.name());
  Serial.print(F(" vers "));
  Serial.println(to.name());
#endif
  from.desactivate();
  pActive = to.activate();
  return true;
}

boolean after(Stat *&pActive,  Stat &from,  Stat &to, unsigned long delay_ms, boolean guard) {  //delay: durée avant le changement d'état, guard : condition de garde, from : état avant, to : état après, pActive : pointeur sur l'état actif
  if (!(from.after(delay_ms) && guard))                                                                   //la condition n'est pas vérifié
    return false;
//cas de la transition franchissable
#if LOG_TRANSITION
  Serial.println();
  Serial.print(F("-> after de "));
  Serial.print(from.name());
  Serial.print(F(" vers "));
  Serial.println(to.name());
#endif
  from.desactivate();
  pActive = to.activate();
  return true;
}

//***********************************
//member definition Timer
//***********************************
Timer::Timer() {
  m_stattime = millis();
  m_stat = false;
}

void Timer::start() {
  m_stattime = millis();
  m_stat = true;
}

void Timer::stop() {
  m_stat = false;
}

bool Timer::stat() {
  return m_stat;
}

bool Timer::after(unsigned long delay_ms) {
  if (!m_stat)  //Etat stop
    return false;

  if ((millis() - m_stattime) < delay_ms)
    return false;
  else {
    return true;
  }
}

//***********************************
//member definition Pulse
//***********************************
Pulse::Pulse() {
  m_every = false;
  m_stattime = millis();
}

bool Pulse::every(unsigned long delay_ms) {
  if (!m_every) {
    m_every = true;
    m_stattime = millis();
    return false;
  } else {
    if ((millis() - m_stattime) < delay_ms) {
      return false;
    } else {
      m_every = false;
      return true;
    }
  }
}

//***********************************
//member definition Event
//***********************************
Event::Event(const char *name) {
  log = false;
  anti_rebounce_time = ANTI_REBOUNCE_TIME;  //Réglage par défaut
  m_stat = LOW;
  m_new_stat = LOW;
  m_rising = false;
  m_falling = false;
  m_pin = 0;
  m_stattime = 0;  //date du derier changement d'état
  m_reverse = false;
  m_name = String(name);
}
void Event::setPin(uint8_t pin, bool interne_pull_up, bool reverse) {
  m_pin = pin;
  m_reverse = interne_pull_up ^ reverse;  //Inversion du résulatat par ou exclusif si la résistance de pull up interne est active ou si inversion explicite V1.5.4
  if (interne_pull_up)
    pinMode(m_pin, INPUT_PULLUP);
  else
    pinMode(m_pin, INPUT);
  //initialisation
  m_stat = (digitalRead(m_pin) ^ m_reverse);  //Inversion du résulatat par ou exclusif si la résistance de pull up interne est active ou si inversion explicite
  m_stattime = millis();
}

bool Event::rising() const{
  return m_rising;
}
bool Event::falling() const{
  return m_falling;
}
bool Event::change() const{
  return (m_falling || m_rising);
}

bool Event::stat() const{
  return m_stat;
}

unsigned long Event::getStattime() {
  return m_stattime;
}

bool Event::update() {  //Retourne false si aucun changement
  m_rising = false;
  m_falling = false;

  if ((millis() - m_stattime) < anti_rebounce_time)  //Anti-rebond
    return false;

  m_new_stat = (digitalRead(m_pin) ^ m_reverse);  //Inversion du résulatat par ou exclusif si la résistance de pull up interne est active
  if (m_new_stat == m_stat)                       //Optimiation V1.6.1
    return false;

  //rising
  //if ((m_new_stat==HIGH)&&(m_stat==LOW)){ //old
  if ((m_new_stat) && (!m_stat)) {  //Optimiation V1.6.1
    m_stat = HIGH;
    m_rising = true;
    m_stattime = millis();
    if (log) {
      Serial.print(F("Event : "));
      Serial.println(m_name + F(" rising at ") + String(m_stattime) + F(" ms."));
    }
    return true;
  }
  //falling
  //if ((m_new_stat==LOW)&&(m_stat==HIGH)){ //old
  if ((!m_new_stat) && (m_stat)) {  //Optimiation V1.6.1
    m_stat = LOW;
    m_falling = true;
    m_stattime = millis();
    if (log) {
      Serial.print(F("Event : "));
      Serial.println(m_name + F(" falling at ") + String(m_stattime) + F(" ms."));
    }
    return true;
  }
  return false;
}


//***********************************
//member definition Button
//***********************************
Button::Button(const char *name)
  : Event(name) {
  m_longPress = false;
  m_longPressEvent = false;
  m_click = false;
  m_clickEvent = false;
  m_long_press_time = LONG_PRESS_TIME;  //Réglage par défaut
}


void Button::setLongPressTime(unsigned long delay_ms) {  //2 secondes par défaut (cf constructeur)
  m_long_press_time = delay_ms;
}

bool Button::longPress() const{
  return (m_longPressEvent && m_longPress);
}

bool Button::longPressFalling() const{
  return (m_longPressEvent && !(m_longPress));
}

bool Button::click() const{
  return m_clickEvent;
}

bool Button::doubleClick() const{
  return m_doubleClickEvent;
}



bool Button::longPressStat() const{
  return m_longPress;
}


bool Button::update() {
  m_rising = false;
  m_falling = false;
  m_longPressEvent = false;
  m_clickEvent = false;
  m_doubleClickEvent = false;

  if ((millis() - m_stattime) < anti_rebounce_time)  //Anti-rebond
    return false;

  m_new_stat = (digitalRead(m_pin) ^ m_reverse);  //Inversion du résulatat par ou exclusif si la résistance de pull up interne est active


  //if (m_new_stat==HIGH) { //old
  if (m_new_stat) {  //optimisation V1.6.1
    //rising
    if (!m_stat) {  //optimisation V1.6.1
      m_stat = HIGH;
      m_rising = true;
      m_stattime = millis();
      if (log) {
        Serial.print(F("Button : "));
        Serial.println(m_name + F(" rising at ") + String(m_stattime) + F(" ms."));
      }
      return true;
    }

    //longPress
    if (((millis() - m_stattime) > m_long_press_time) && !m_longPress) {
      if (log) {
        Serial.print(F("Button : "));
        Serial.println(m_name + F(" longPress at ") + String(m_stattime) + " ms.");
      }
      m_longPress = true;
      m_longPressEvent = true;
      return true;
    }
  }
  //if (m_new_stat==LOW) {	 //old
  if (!m_new_stat) {  //optimisation V1.6.1
    //falling
    if (m_stat) {  //optimisation V1.6.1
      m_stat = LOW;
      m_falling = true;
      if (m_longPress) {
        m_longPress = false;
        m_longPressEvent = true;
      }
      if ((millis() - m_stattime) < CLICK_TIME) {
        if (m_click) {
          //doubleClic
          if (log) {
            Serial.print(F("Button : "));
            Serial.println(m_name + F(" doubleClick at ") + String(m_stattime) + F(" ms."));
          }
          m_doubleClickEvent = true;
          m_click = false;
        } else
          m_click = true;
      }
      m_stattime = millis();
      if (log) {
        Serial.print(F("Button : "));
        Serial.println(m_name + F(" falling at ") + String(m_stattime) + F(" ms."));
      }
    }

    //Click
    if ((millis() - m_stattime) > CLICK_TIME) {
      if (m_click) {
        if (log) {
          Serial.print(F("Button : "));
          Serial.println(m_name + F(" click at ") + String(m_stattime) + F(" ms."));
        }
        m_clickEvent = true;
        m_click = false;
        return true;
      }
    }
    return m_falling;
  }
  return false;
}

//***********************************
//member definition EventFiltered
//***********************************
EventFiltered::EventFiltered(const char *name)
  : Event(name) {
  m_filtered_time = FILTERED_TIME;  //Réglage par défaut
  m_wait = false;
  m_wait_time = 0;
}


void EventFiltered::setFilteredTime(unsigned long delay_ms) {  //0.5 secondes par défaut (cf constructeur)
  m_filtered_time = delay_ms;
}

bool EventFiltered::update() {  //Retourne false si aucun changement
  m_rising = false;
  m_falling = false;

  if ((millis() - m_stattime) < anti_rebounce_time)  //Anti-rebond
    return false;

  m_new_stat = (digitalRead(m_pin) ^ m_reverse);  //Inversion du résulatat par ou exclusif si la résistance de pull up interne est active
  if (m_new_stat == m_stat) {
    if (m_wait) {
      m_wait = false;  //Annulation de l'évenement
      if (log) {
        Serial.print(F("EventFiltered : "));
        Serial.println(m_name + F(" canceled event ") + String(millis()) + F(" ms."));
      }
    }
    return false;
  }

  //rising
  //if ((m_new_stat==HIGH)&&(m_stat==LOW)){ //old
  if ((m_new_stat) && (!m_stat)) {  //Optimiation V1.6.1
    if (!m_wait) {                  //_wait == false
      m_wait = true;
      m_wait_time = millis();  //Initialisation de l'attente
      return false;
    } else  //_wait == true
      if (millis() - m_wait_time < m_filtered_time)
        return false;
      else {
        m_stat = HIGH;
        m_rising = true;
        m_stattime = millis();
        m_wait = false;
        if (log) {
          Serial.print(F("EventFiltered : "));
          Serial.println(m_name + F(" rising at ") + String(m_stattime) + F(" ms."));
        }
        return true;
      }
  }


  //falling
  //if ((m_new_stat==LOW)&&(m_stat==HIGH)){ //old
  if ((!m_new_stat) && (m_stat)) {  //Optimiation V1.6.1
    if (!m_wait) {                  //_wait == false
      m_wait = true;
      m_wait_time = millis();  //Initialisation de l'attente
      return false;
    } else  //_wait == true
      if (millis() - m_wait_time < m_filtered_time)
        return false;
      else {
        m_stat = LOW;
        m_falling = true;
        m_stattime = millis();
        m_wait = false;
        if (log) {
          Serial.print(F("EventFiltered : "));
          Serial.println(m_name + F(" falling at ") + String(m_stattime) + F(" ms."));
        }
        return true;
      }
  }
  return false;
}

//***********************************
//Bistable
//***********************************

Bistable::Bistable(const char *name1, const char *name2) {
  m_pin1 = 0;
  m_pin2 = 0;
  m_name1 = String(name1);
  m_name2 = String(name2);
  log = false;
  m_stat1 = false;
  m_stat2 = false;
}
void Bistable::setPin(uint8_t pin1, uint8_t pin2, bool reverse) {
  m_pin1 = pin1;
  m_pin2 = pin2;
  m_reverse = reverse;
  pinMode(m_pin1, OUTPUT);
  pinMode(m_pin2, OUTPUT);
  if (m_reverse) {
    digitalWrite(m_pin1, HIGH);
    digitalWrite(m_pin2, HIGH);
  }
}

void Bistable::set1(bool stat) {
  if ((stat) && (digitalRead(m_pin2) ^ m_reverse)) {
    Serial.print(F("Bistable : Attention, activation de la sortie 1 sans que la sortie 2 soit désactive"));
    digitalWrite(m_pin2, m_reverse);
    delay(100);
  }

  digitalWrite(m_pin1, stat ^ m_reverse);
  m_stat1 = stat;

  if (log)
    if (stat)
      Serial.println("-->" + m_name1 + F(" ON at ") + String(millis()) + F(" ms."));
    else
      Serial.println("-->" + m_name1 + F(" OFF at ") + String(millis()) + F(" ms."));
}
void Bistable::set2(bool stat) {
  if ((stat) && (digitalRead(m_pin1) ^ m_reverse)) {
    Serial.print(F("Bistable : Attention, activation de la sortie 2 sans que la sortie 1 soit désactive"));
    digitalWrite(m_pin1, m_reverse);
    delay(100);
  }

  digitalWrite(m_pin2, stat ^ m_reverse);
  m_stat2 = stat;

  if (log)
    if (stat)
      Serial.println("-->" + m_name2 + " ON at " + String(millis()) + F(" ms."));
    else
      Serial.println("-->" + m_name2 + " OFF at " + String(millis()) + F(" ms."));
}

bool Bistable::stat1() const{
  return m_stat1;
}

bool Bistable::stat2() const{
  return m_stat2;
}

//***********************************
//DigitalOut
//***********************************

DigitalOut::DigitalOut(const char *name) {
  m_pin = 0;
  m_name = String(name);
  m_stattime = 0;
  m_stat = false;
  m_flash = false;
  m_flash_period = 0;
  m_delay = 0;
  m_change = false;
  m_end_of_timer = false;
  log = false;
}
void DigitalOut::setPin(uint8_t pin, bool reverse) {
  m_pin = pin;
  m_reverse = reverse;
  pinMode(m_pin, OUTPUT);
  if (m_reverse) {
    digitalWrite(m_pin, HIGH);
  }
}
void DigitalOut::on(unsigned long delay_on) {
  m_delay = delay_on;
  digitalWrite(m_pin, !m_reverse);
  m_flash = false;
  m_stattime = millis();
  m_stat = true;
  m_change = true;
  if (log) {
    Serial.print(F("DigitalOut : "));
    if (m_delay == 0)
      Serial.println(m_name + F(" ON at ") + String(m_stattime) + F(" ms."));
    else
      Serial.println(m_name + F(" ON (for ") + String(m_delay) + F(" ms) at ") + String(m_stattime) + F(" ms."));
  }
}

void DigitalOut::off(unsigned long delay_off) {
  m_delay = delay_off;
  digitalWrite(m_pin, m_reverse);
  m_flash = false;
  m_stattime = millis();
  m_stat = false;
  m_change = true;
  if (log) {
    Serial.print(F("DigitalOut : "));
    if (m_delay == 0)
      Serial.println(m_name + F(" OFF at ") + String(m_stattime) + F(" ms."));
    else
      Serial.println(m_name + F(" OFF (for ") + String(m_delay) + F(" ms) at ") + String(m_stattime) + F(" ms."));
  }
}

void DigitalOut::set(bool on_off) {  //mise à 1 ou à 0
  m_delay = 0;
  digitalWrite(m_pin, on_off ^ m_reverse);
  m_flash = false;
  m_stattime = millis();
  m_stat = on_off;
  m_change = true;
  if (log) {
    Serial.print(F("DigitalOut : "));
    if (on_off)
      Serial.println(m_name + F(" ON at ") + String(m_stattime) + F(" ms."));
    else
      Serial.println(m_name + F(" OFF at ") + String(m_stattime) + F(" ms."));
  }
}

void DigitalOut::toggle() {
  m_delay = 0;
  m_stat = !m_stat;
  digitalWrite(m_pin, m_stat ^ m_reverse);
  m_flash = false;
  m_stattime = millis();
  m_change = true;
  if (log) {
    Serial.print(F("DigitalOut : "));
    if (m_stat)
      Serial.println(m_name + F(" ON at ") + String(m_stattime) + F(" ms."));
    else
      Serial.println(m_name + F(" OFF at ") + String(m_stattime) + F(" ms."));
  }
}

void DigitalOut::flash(unsigned long flash_period) {
  if (log && !m_flash) {
    Serial.print(F("DigitalOut : "));
    Serial.println(m_name + F(" flash on (Flash period :") + String(flash_period) + F(" ms) at ") + String(m_stattime) + F(" ms."));
  }
  m_flash = true;
  m_flash_period = flash_period;
  m_delay = 0;
}

bool DigitalOut::change() const{
  return m_change;
}


bool DigitalOut::rising() const{  //V1.6.3
  return (m_change && m_stat);
}

bool DigitalOut::falling() const{  //V1.6.3
  return (m_change && (!m_stat));
}

bool DigitalOut::endOfTimer() const{  //V1.6.3 //V1.7.0
  return m_end_of_timer;
}

bool DigitalOut::update() {
  bool change = m_change;  //V.1.6.3
  m_change = false;
  m_end_of_timer = false;
  if (m_flash) {
    if ((millis() - m_stattime) > (m_flash_period / 2)) {
      m_stat = !m_stat;
      digitalWrite(m_pin, m_stat ^ m_reverse);
      m_stattime = millis();
      m_change = true;  //V.1.6.3
    }
  }
  if (m_delay != 0) {
    if ((millis() - m_stattime) > m_delay) {
      m_end_of_timer = true;
      if (m_stat)
        off();
      else
        on();
      //m_change = true ; //V.1.6.3 déjà dans on et off
    }
  }
  return (change || m_change);  //V.1.6.3
}

bool DigitalOut::stat() const{
  return m_stat;
};