#include <Arduino.h>
#include "obj_stat.h"

//******************************************
//Portail GTI Version GTI4 du 08 06 2020
//A DUGAS / CPGE PT ATS - Lycee Livet
//Prévu pour fonctionner sur une carte MEGA 2560
//Dans le menu outil Outils :
// * Choisir Type de carte : Arduino Mega 2560
//Dans le moniteur série :
// * Choisir une vitesse de 9600 bauds
//******************************************

//******************************************
//Déclaration hardware
//******************************************
//Entrées
#define BP_OUVERTURE A6          //Module Grove
#define BP_FERMETURE A8          //Module Grove
#define BP_OUVERTURE_PIETON A10  //Module Grove

#define FC_OUVERTURE 4  //24V via opto-coupleur
#define FC_FERMETURE 6  //24V via opto-coupleur

////////////////////////////////////////////////
//ATTENTION : Mettre l'une des 2 lignes en commentaire
//suivant qu'on est sur la platine de test ou sur le
//système

//#define OPTO         2   //Bouton grove sur platine test
#define OPTO 5  //24V via opto-coupleur sur système

//#define INDUCTIF         12   //Bouton grove sur platine test
#define INDUCTIF 7  //24V via opto-coupleur sur système
///////////////////////////////////////////////

//Sorties
#define LED_MVT A0     //Module Grove (Rouge)
#define LED_OUVERT A2  //Module Grove (Verte)
#define LED_FERME A4   //Module Grove (Rouge)
#define OUVERTURE 8    //24V via relais
#define FERMETURE 10   //24V via relais

//******************************************
//Déclaration des états
//******************************************
Stat E_ferme("Ferme", 100);
Stat E_en_ouverture("En cours d'ouverture", 110);
Stat E_en_ouverture_pieton("En cours d'ouverture partielle", 150);
Stat E_ouvert("Ouvert", 120);
Stat E_en_fermeture("En cours de fermeture", 130);
Stat E_en_pause("En pause de 2 secondes", 140);

Stat* pEtat_actif;  //pointeur vers l'état actif

//******************************************
//Déclaration des évènements
//******************************************
Button bp_ouverture;
Button bp_fermeture;
Button bp_ouverture_pieton;
Event fc_ouverture;
Event fc_fermeture;
Event opto("optique");
Event inductif("inductif");

//******************************************
//Déclaration des sorties
//******************************************
Bistable moteur("Relais d'ouverture", "Relais de fermeture");
DigitalOut led_mvt("Led de d'avertissement");
DigitalOut led_fc_ouverture("Led fc ouverture");
DigitalOut led_fc_fermeture("Led fc fermeture");

//******************************************
//Déclaration des variables
//******************************************
unsigned long compteur;
Timer timer;

//******************************************
//Setup
//******************************************
void setup() {
  //******************************************
  //Setup hardware
  //******************************************
  //Entrées
  bp_ouverture.setPin(BP_OUVERTURE);
  bp_fermeture.setPin(BP_FERMETURE);
  bp_ouverture_pieton.setPin(BP_OUVERTURE_PIETON);
  fc_ouverture.setPin(FC_OUVERTURE);
  fc_fermeture.setPin(FC_FERMETURE);
  opto.setPin(OPTO);
  inductif.setPin(INDUCTIF);

  //Sorties
  led_mvt.setPin(LED_MVT);
  led_fc_ouverture.setPin(LED_OUVERT);
  led_fc_fermeture.setPin(LED_FERME);
  moteur.setPin(OUVERTURE, FERMETURE);

  //******************************************
  //Port série
  //******************************************
  Serial.begin(9600);
  //******************************************
  //Initialisation des évènements
  //******************************************
  bp_ouverture.update();
  bp_fermeture.update();
  bp_ouverture_pieton.update();

  fc_ouverture.update();
  fc_fermeture.update();
  opto.update();
  inductif.update();
  //******************************************
  //Etat initial
  //******************************************
  if (!fc_fermeture.stat())
    pEtat_actif = E_ouvert.activate();
  else
    pEtat_actif = E_ferme.activate();
  compteur = 0;
  timer.start();
  //******************************************
  //Activation des log
  //******************************************
  led_mvt.log = false;
  led_fc_ouverture.log = false;
  led_fc_fermeture.log = false;
  moteur.log = false;
  opto.log = false;
  inductif.log = true;
}
//******************************************
//loop
//******************************************
void loop() {
  //******************************************
  //Evolution des états
  //******************************************

  //Ouvert->En Fermeture
  transition(pEtat_actif, E_ouvert, E_en_fermeture, bp_fermeture.rising());

  //En Fermeture->Fermé
  transition(pEtat_actif, E_en_fermeture, E_ferme, fc_fermeture.rising());

  //En Fermeture->En pause
  transition(pEtat_actif, E_en_fermeture, E_en_pause, opto.rising());

  //En pause->En ouverture
  after(pEtat_actif, E_en_pause, E_en_ouverture, 2000);

  //Fermé->En ouverture
  transition(pEtat_actif, E_ferme, E_en_ouverture, bp_ouverture.rising());

  //Fermé->En ouverture pieton
  transition(pEtat_actif, E_ferme, E_en_ouverture_pieton, bp_ouverture_pieton.rising());

  //En ouverture->Ouvert
  transition(pEtat_actif, E_en_ouverture, E_ouvert, fc_ouverture.rising());

  //En ouverture pieton->Ouvert
  transition(pEtat_actif, E_en_ouverture_pieton, E_ouvert, fc_ouverture.rising());
  when(pEtat_actif, E_en_ouverture_pieton, E_ouvert, (compteur == 5));
  //******************************************
  //Actions pendant les états
  //******************************************
  //Exit
  if (E_ouvert.exit_()) {
    led_fc_ouverture.off();
  }
  if (E_ferme.exit_()) {
    led_fc_fermeture.off();
  }
  //Do
  if (E_en_ouverture.do_()) {
    led_mvt.flash(500);
    if (inductif.rising() && timer.after(1000)) {
      compteur++;
      timer.start();  //on redémarre le timer
      Serial.println("   *Compteur: " + String(compteur) + " tours");
    }
  }
  if (E_en_ouverture_pieton.do_()) {
    led_mvt.flash(500);
    if (inductif.rising() && timer.after(1000)) {
      compteur++;
      timer.start();  //on redémarre le timer
      Serial.println("   *Compteur: " + String(compteur) + " tours");
    }
  }
  if (E_en_fermeture.do_()) {
    led_mvt.flash(500);
    if (inductif.rising() && timer.after(1000)) {
      compteur--;
      timer.start();  //on redémarre le timer
      Serial.println("   *Compteur: " + String(compteur) + " tours");
    }
  }
  //Entry
  if (E_en_ouverture.entry_()) {
    moteur.set1(HIGH);
  }
  if (E_en_ouverture_pieton.entry_()) {
    moteur.set1(HIGH);
  }

  if (E_en_fermeture.entry_()) {
    moteur.set2(HIGH);
  }
  if (E_ouvert.entry_()) {
    moteur.set1(LOW);
    led_fc_ouverture.on();
    led_mvt.off();
  }
  if (E_ferme.entry_()) {
    moteur.set2(LOW);
    led_fc_fermeture.on();
    led_mvt.off();
    compteur = 0;
  }
  if (E_en_pause.entry_()) {
    moteur.set2(LOW);
    led_mvt.on();
  }
  //******************************************
  //Mise à jour des états
  //******************************************
  E_ouvert.update();
  E_ferme.update();
  E_en_ouverture.update();
  E_en_ouverture_pieton.update();
  E_en_fermeture.update();
  E_en_pause.update();

  //******************************************
  //Mise à jour des évenements
  //******************************************
  bp_ouverture.update();
  bp_fermeture.update();
  bp_ouverture_pieton.update();
  fc_ouverture.update();
  fc_fermeture.update();
  opto.update();
  inductif.update();
  //******************************************
  //Mise à jour des sorties
  //******************************************
  led_mvt.update();
  led_fc_ouverture.update();
  led_fc_fermeture.update();
}